﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.DataPreprocessing {
  [Item("PreprocessingDataColumn", "")]
  [StorableClass]
  public abstract class PreprocessingDataColumn : NamedItem {

    #region Constructor, Cloning & Persistence
    protected PreprocessingDataColumn() {
    }
    protected PreprocessingDataColumn(string name)
      : base(name) {
    }

    protected PreprocessingDataColumn(PreprocessingDataColumn original, Cloner cloner)
      : base(original, cloner) {
    }

    [StorableConstructor]
    protected PreprocessingDataColumn(bool deserializing)
      : base(deserializing) { }
    #endregion

    public abstract int Length { get; }
    public abstract Type GetValueType();
    public virtual bool IsType<T>() { return GetValueType() == typeof(T); }

    public abstract bool IsValidValue(int index);


    #region String Handling
    public abstract bool Validate(string value, out string errorMessage);
    public abstract string GetValue(int index);
    public abstract bool SetValue(string value, int index);
    #endregion
  }

  [Item("PreprocessingDataColumn", "")]
  [StorableClass]
  public abstract class PreprocessingDataColumn<TValue, TDistance> : PreprocessingDataColumn
    where TValue : class, IComparable<TValue> {

    #region Constructor, Cloning & Persistence
    protected PreprocessingDataColumn()
      : this(string.Empty, Enumerable.Empty<TValue>()) { }
    protected PreprocessingDataColumn(string name, IEnumerable<TValue> values)
      : base(name) {
      Values = new List<TValue>(values);
    }

    protected PreprocessingDataColumn(PreprocessingDataColumn<TValue, TDistance> original, Cloner cloner)
      : base(original, cloner) {
      Values = new List<TValue>(original.Values);
    }

    [StorableConstructor]
    protected PreprocessingDataColumn(bool deserializing)
      : base(deserializing) { }
    #endregion

    [Storable]
    public List<TValue> Values { get; private set; }
    public IEnumerable<TValue> ValidValues {
      get { return Values.Where(IsValidValue); }
    }

    public override Type GetValueType() {
      return typeof(TValue);
    }

    public override int Length {
      get { return Values.Count; }
    }

    public TValue this[int index] {
      get { return Values[index]; }
      set { Values[index] = value; }
    }

    public virtual bool IsValidValue(TValue value) { return true; }
    public override bool IsValidValue(int index) {
      return IsValidValue(Values[index]);
    }

    #region Statistics
    public virtual TValue GetMin() { return Values.Min(); }
    public virtual TValue GetMax() { return Values.Max(); }
    public abstract TDistance GetRange();
    public abstract TValue GetMean();
    public virtual TValue GetMedian() { return Values.Quantile(0.5); }
    public virtual TValue GetMode() { return Values.GroupBy(x => x).OrderByDescending(g => g.Count()).Select(g => g.Key).First(); }
    public abstract TDistance GetStandardDeviation();
    public abstract TDistance GetVariance();
    public virtual TValue GetQuantile(double alpha) { return Values.Quantile(alpha); }
    public virtual int GetDistinctValues() { return Values.GroupBy(x => x).Count(); }
    public virtual int GetNumberOfMissingValues() { return Values.Count(IsValidValue); }
    #endregion
  }

  [Item("NullablePreprocessingDataColumn", "")]
  [StorableClass]
  public abstract class NullablePreprocessingDataColumn<TValue, TDistance> : PreprocessingDataColumn
    where TValue : struct, IComparable<TValue> {

    #region Constructor, Cloning & Persistence
    protected NullablePreprocessingDataColumn()
      : this(string.Empty, Enumerable.Empty<TValue?>()) { }
    protected NullablePreprocessingDataColumn(string name, IEnumerable<TValue> values)
      : this(name, values.Select(x => (TValue?)x)) { }
    protected NullablePreprocessingDataColumn(string name, IEnumerable<TValue?> values)
      : base(name) {
      Values = new List<TValue?>(values);
    }

    protected NullablePreprocessingDataColumn(NullablePreprocessingDataColumn<TValue, TDistance> original, Cloner cloner)
      : base(original, cloner) {
      Values = new List<TValue?>(original.Values);
    }

    [StorableConstructor]
    protected NullablePreprocessingDataColumn(bool deserializing)
      : base(deserializing) { }
    #endregion

    [Storable]
    internal List<TValue?> Values { get; private set; }
    protected IEnumerable<TValue> ValidValues {
      get { return Values.Where(x => x.HasValue && IsValidValue(x.Value)).Select(x => x.Value); }
    }

    public override Type GetValueType() {
      return typeof(TValue);
    }

    public override int Length {
      get { return Values.Count; }
    }

    public TValue? this[int index] {
      get { return Values[index]; }
      set { Values[index] = value; }
    }

    public virtual bool IsValidValue(TValue value) { return true; }
    public override bool IsValidValue(int index) {
      var value = Values[index];
      return value.HasValue && IsValidValue(value.Value);
    }

    #region Statistics
    public virtual TValue GetMin() { return ValidValues.Min(); }
    public virtual TValue GetMax() { return ValidValues.Max(); }
    public abstract TDistance GetRange();
    public abstract TValue GetMean();
    public virtual TValue GetMedian() { return ValidValues.Quantile(0.5); }
    public virtual TValue GetMode() { return ValidValues.GroupBy(x => x).OrderByDescending(g => g.Count()).Select(g => g.Key).First(); }
    public abstract TDistance GetStandardDeviation();
    public abstract TDistance GetVariance();
    public virtual TValue GetQuantile(double alpha) { return ValidValues.Quantile(alpha); }
    public virtual int GetDistinctValues() { return ValidValues.GroupBy(x => x).Count(); }
    public virtual int GetNumberOfMissingValues() { return Values.Count - ValidValues.Count(); }
    #endregion
  }
}